#!/usr/bin/env python

from distutils.core import setup, Extension
import os
import glob

name = "python-inotify"
version = "0.0.1"

module1 = Extension('_inotify',
                    sources = ['inotifymodule.c'])

setup(name=name,
	version=version,
	description="The Linux inotify abstraction",
	author="Manuel Amador (Rudd-O)",
	author_email="rudd-o@amautacorp.com",
	url="http://www.amautacorp.com/staff/Rudd-O/projects/search-services/",
	py_modules = ["inotify"],
	ext_modules = [module1],
	)
